module Main where

import Lib
import System.Random
import Data.Char

-- The 'Main' module is the top-level module for the stand-alone application.

-- The 'main' action of the 'Main' module is the entry-point for the
-- application. When GHC produces an executable file, that executable will have
-- the effect of *running* 'main'. The *only* actions that will run in a Haskell
-- program are those that are called (transitively) from 'main'.

-- Note that 'main' always has the type IO a, for some type a. The type includes
-- IO because it will perform IO effects like console interaction,
-- reading/writing files, network communication, etc. The type of value returned
-- by the 'main' action here is the boring (), because the return value doesn't
-- matter; there is no one to return to once 'main' is done.
main :: IO Int
main = do
  putStrLn "Welcome to the guessing game!"
  forever (playRound (1, 100))

-- Play a round of the guessing game, given a range of numbers to choose as the
-- secret.
playRound :: (Int, Int) -> IO ()
playRound range = do
  secret <- randomRIO range
  let (low, high) = range
  putStrLn ("I've chosen a number between " ++ show low ++ " and " ++ show high ++ ".")
  repeatUntilTrue (makeAGuess secret)
  return ()
-- The function randomRIO can be used with the specialized type
--
--     randomRIO :: (Int, Int) -> IO Int
--
-- which takes a range (low, high) :: (Int, Int) as an argument and performs an
-- IO action to generate a random number in between low and high, inclusively.
-- In other words, the random number generated by randomRIO (low, high) will be
-- some 'r' such that low <= r <= high.

-- Have the user make a guess at the secret number. Returns True if the guess
-- was correct, False otherwise.
makeAGuess :: Int -> IO Bool
makeAGuess secret = do
  answer <- prompt "What's my number?"
  let guess    = read answer :: Int
      sensible = all isDigit answer
  case (sensible, compare guess secret) of
    (False, _) -> do putStrLn "That's not even a number!"
                     return False
    (True, LT) -> do putStrLn "That's too low."
                     return False
    (True, GT) -> do putStrLn "That's too high."
                     return False
    (True, EQ) -> do putStrLn "That's it!"
                     return True
-- Note the new form of expression here.  There is a 'case' expression which
-- does pattern matching on the expression between the 'case' and the 'of'. A
-- 'case' can be used on any data type, as well as numbers and Strings. 'case'
-- does not add any new functionality that we couldn't do with pattern matching
-- in function declarations. But it is useful here in the middle of a 'do'
-- because it is an *expression*.

-- Also note that makeAGuess does some sanity checking on the user input, which
-- may be totally bogus. The user might not type in a string of characters like
-- "12fsg5" which does not sensibly represent a number. read "12fsg5" of type
-- Int will fail, so we should first check that what the user said makes sense,
-- by checking that it is a String of only digit characters, before calling a
-- fallible function like 'read'.
